import chimera, numpy as np
from StructBio.Scenographics.labels import LabelGroups
from StructBio.Scenographics.solids import Solids
from StructBio.Scenographics.parametricSurfaces import FrenetFrameSurface

#===============================================================================
# PieChart CLASS
#===============================================================================
class PieChart(object):
    def __init__(self, data_L):

        self.data_L = data_L
        self.numPieces = len(self.data_L)

        # Do computations necessary to define shape of a pie piece:        
        self.data_L.sort()
        self.amounts_L = [entry[0] for entry in self.data_L]
        self.titles_L = [entry[1] for entry in self.data_L]

        sumAmts = sum(self.amounts_L)
        percentAmts_L = [100. * x/sumAmts for x in self.amounts_L]
        self.amounts_L = [2. * np.pi * x/sumAmts for x in self.amounts_L]

        # Get start and stop theta values for each pie piece:
        thetaIntervals_L = []
        startTheta = -np.pi/4.
        for val in self.amounts_L:
            endTheta = startTheta + val
            thetaIntervals_L.append((startTheta, endTheta))
            startTheta = endTheta

        # Define pie profile (a half ellipse) to be used as the base curve):
        a = 10.0
        b = 1.0
        incr = np.pi/64
        seq = np.arange(-np.pi/2, np.pi/2 + incr, incr)
        basePts_L = []
        for theta in seq:
            basePts_L.append((a * np.cos(theta) - a, b * np.sin(theta)))


        self.colors_L = []
        self.transparencies_L = []
        for i in range(self.numPieces):
            self.colors_L.append((float(i)/self.numPieces, np.random.rand()/2., np.random.rand()))
            self.transparencies_L.append(1.0)
            
        self.piePieces_L = []
        # Build pieces of pie:
        for i in range(self.numPieces):
            self.piePieces_L.append(PiePiece(percentAmts_L[i], self.titles_L[i], basePts_L,
                                             thetaIntervals_L[i][0], thetaIntervals_L[i][1],
                                             self.colors_L[i], transparency = self.transparencies_L[i]))

        chimera.runCommand("set bg_color white")
        chimera.runCommand("focus")



#===============================================================================
# PiePiece CLASS
#===============================================================================
class PiePiece(FrenetFrameSurface):
    def __init__(self, percentage, title, baseCurvePts_L, thetaMin, thetaMax, color, transparency = 1.0):

        self.arrows = Solids(title + " Arrow")
        self.pieLabels = LabelGroups("Label for " + title)
        self.pieLabels.addLabelGroup(title + " Label")
        
        spinePts_L = []
        rotMats_L = []
        incr = 2.0*np.pi/256
        for u in np.arange(thetaMin, thetaMax + incr, incr, dtype = float):
            pt = 10.0 * np.array([np.cos(u), np.sin(u), 0.0])
            spinePts_L.append(pt)
            rotMats_L.append(np.array([[np.cos(u), np.sin(u), 0.],
                                       [0.0, 0.0, 1.0],
                                       [np.sin(u), -np.cos(u), 0.]]).transpose())

        # Push pie pieces away from pie center:
        middlePathPt = spinePts_L[len(spinePts_L)/2]
        spinePts_L = [p + middlePathPt/20. for p in spinePts_L]

        FrenetFrameSurface.__init__(self, baseCurvePts_L, spinePts_L, rotMats_L, [1.], [1.])
        self.modelname = title + "Surface"
        self.displaySurface(color, transparency = transparency, closedBase = True,
                            startCap = True, endCap = True, closedSpine = False)

        # Add arrow:
        self.arrows.addArrow(middlePathPt * 1.5, middlePathPt * 1.1, .05, (0.2, 0.2, 0.2))

        # Add label:
        self.pieLabels.addLabel(title + " Label", " %5.1f %%" % percentage + "  " + title, middlePathPt * 1.55)
                    
        self.arrows.display()
        self.pieLabels.setLabels(title + " Label", "red")
        self.pieLabels.showLabels()                 
    

#===============================================================================
# GUI imports:
import Tkinter as tk, tkColorChooser, tkFileDialog
from chimera.baseDialog import ModelessDialog

class PiePlotGUI(ModelessDialog):
    buttons = ("Close")
    help = False    # Leave this out to get a Help button.
    title = "Pie Chart"

    def fillInUI(self, parent):
        #-------------------------------------------------------------------------
        # GUI state variables:
        self.data_L = []    # List of pairs. Each pair is a tuple: (data value, title).
        
        #-------------------------------------------------------------------------
        # FETCH FRAME
        fetchData_frame = tk.LabelFrame(parent, text =  " Generate Pie Chart ", bd = 2, relief = tk.RIDGE)
        fetchData_frame.grid(row = 0, column = 0, padx = 4, pady = 4, sticky = tk.W + tk.E)

        fetchData_text = tk.Label(fetchData_frame, text =  " Input data:   ")
        fetchData_text.grid(row = 0, column = 0, pady = 4, sticky = tk.W)

        self.fetchData_btn = tk.Button(fetchData_frame, text = " Fetch ",
                                      command = self.fetchData_btn_Func)
        self.fetchData_btn.grid(row = 0, column = 1, padx = 4, pady = 4)

        #-------------------------------------------------------------------------
        # SELECT PIE FRAME
        self.selectPie_frame = tk.LabelFrame(parent, text =  " Select Pie Piece ", bd = 2, relief = tk.RIDGE)
        self.selectPie_frame.grid(row = 1, column = 0, padx = 4, pady = 4, sticky = tk.W + tk.E)


    #=============================================================================
    # CALLBACK functions:
    #-------------------------------------------------------------------------
    def fetchData_btn_Func(self):

        # Launch dialog to get path of PDB ID list:
        filePathName = tkFileDialog.askopenfilename()
        print "Path name for data file is: ", filePathName

        # Read file containing list of count and title pairs:
        listFileHandle = file(filePathName)
        listFileHandle.readline()    # Skip first line
        listLine = listFileHandle.readline()
        while listLine:
            listLineEntries_L = listLine.split()
            title = " ".join(listLineEntries_L[1: len(listLineEntries_L)])
            self.data_L.append((float(listLineEntries_L[0]), title))
            listLine = listFileHandle.readline()
        listFileHandle.close()  # Close the list file.

        self.fetchData_btn.configure(state = tk.DISABLED)
        
        self.pieChart = PieChart(self.data_L)

        # Fill in the select pie frame:
        # For best placement, lay out buttons in a subframe (no ridge) for better spacing.
        rBtns_frame = tk.Frame(self.selectPie_frame, relief = tk.FLAT)
        rBtns_frame.grid(row = 0, column = 0, columnspan = 5, padx = 4, pady = 4, sticky = tk.W + tk.E)
        
        self.selectedPiece_Rbtn_Result = tk.IntVar()
        for i in range(self.pieChart.numPieces):
            rBtn = tk.Radiobutton(rBtns_frame, text = self.pieChart.titles_L[i] + "    ", variable = self.selectedPiece_Rbtn_Result,
                                  value = i, command = self.choosePiece_Rbtns_Func)
            rBtn.grid(row = i/4, column = i%4, sticky = tk.W)
        self.selectedPiece_Rbtn_Result.set(0) # This button starts off selected.


        tk.Label(self.selectPie_frame, text =  " ").grid(row = 20, column = 0, pady = 4, sticky = tk.W) # Spacer
        
        # Color button:
        tk.Label(self.selectPie_frame, text =  " Change color:  ").grid(row = 21, column = 0, pady = 4, sticky = tk.E)
        self.colorBtn = tk.Button(self.selectPie_frame, text = "", width = 5, height = 1, command = self.changeColor_btn_Func)
        self.colorBtn.configure(bg = colorDec2Hex(self.pieChart.colors_L[0]))
        self.colorBtn.grid(row = 21, column = 1, padx = 4, pady = 4, sticky = tk.W)

        # Transparency:
        self.transparency_scaleVar = tk.DoubleVar()
        transparency_text = tk.Label(self.selectPie_frame, text =  " Transparency:  ")
        transparency_text.grid(row = 21, column = 2, pady = 4, sticky = tk.E)
        self.transparency_scale = tk.Scale(self.selectPie_frame, variable = self.transparency_scaleVar,
                                           orient = tk.HORIZONTAL, length = 100, from_ = 0.0, to = 1.0, resolution = 0.1,
                                           troughcolor = "white")
        self.transparency_scale.grid(row = 21, column = 3, columnspan = 3, sticky = tk.W)
        self.transparency_scale.set(1.0)
        
        accept_btn = tk.Button(self.selectPie_frame, text = "Accept", command = self.accept_btn_Func)
        accept_btn.grid(row = 21, column = 6, columnspan = 2, padx = 4, pady = 4, sticky = tk.W)


    #-------------------------------------------------------------------------
    def choosePiece_Rbtns_Func(self):
        pieceNum = self.selectedPiece_Rbtn_Result.get()
        self.colorBtn.configure(bg = colorDec2Hex(self.pieChart.colors_L[pieceNum]))
        self.transparency_scale.set(self.pieChart.transparencies_L[pieceNum])
        
    #-------------------------------------------------------------------------
    def changeColor_btn_Func(self):
        pieceNum = self.selectedPiece_Rbtn_Result.get()
        (rgb, hx) = tkColorChooser.askcolor(initialcolor = colorDec2Hex(self.pieChart.colors_L[pieceNum]), title = "Choose a color")
        self.pieChart.colors_L[pieceNum] = (rgb[0]/255., rgb[1]/255., rgb[2]/255.)
        self.colorBtn.configure(bg = hx)
        self.accept_btn_Func()
        
    #-------------------------------------------------------------------------        
    def accept_btn_Func(self):
        pieceNum = self.selectedPiece_Rbtn_Result.get()
        self.pieChart.transparencies_L[pieceNum] = self.transparency_scaleVar.get()
        self.pieChart.piePieces_L[pieceNum].removeSurface()
        newColor = self.pieChart.colors_L[pieceNum]
        newTrans = self.pieChart.transparencies_L[pieceNum]
        self.pieChart.piePieces_L[pieceNum].displaySurface(newColor, transparency = newTrans, closedBase = True,
                                                           startCap = True, endCap = True, closedSpine = False)


#-------------------------------------------------------------------------
def colorDec2Hex(c_T):
    return "#%02x%02x%02x" % (int(c_T[0]*255), int(c_T[1]*255), int(c_T[2]*255))
        
#===============================================================================
# Mainline:

piePlotGUI = PiePlotGUI()


















    
