import chimera
import numpy as np
from StructBio.Utilities.shellNeighbors import Shell
import matplotlib.pyplot as plot
from matplotlib.colors import ListedColormap

def buildDistMapArray(n, z_D):
    retVal = np.ones((n,n), float)
    for tup in z_D.keys():
        #if tup[0] > tup[1]: continue   # Use this to get upper triange only.
        retVal[tup[0],tup[1]] = z_D[tup]
    return retVal


def generate2Dplot(zG_A, plotTitle):
    # zG_A should be a square array holding values in the range [0.0, 1.0].
    
    # [0,.5) black;  [.5, 1.] white:
    #cmap = ListedColormap(['#000000', '#FFFFFF'])

    # [0., 0.3333] black;   [.3333, .6666] gray;  [.6666, 1.0] white:
    #cmap = ListedColormap(['#000000', '#888888', '#FFFFFF'])

    # [0., 0.2] black; [.2, .4] dark gray;  [.4, .6] gray;  [.6, .8] light gray;  [.8, 1.] white
    cmap = ListedColormap(['#000000', '#444444', '#888888', '#CCCCCC', '#FFFFFF'])
    
    n = int(zG_A.shape[0])
    xSeq = np.arange(0, n+1, 1)
    ySeq = np.arange(0, n+1, 1)
    #ySeq = np.arange(n, -1, -1)  # Use this to have the main diagonal head to the "north west".

    xG, yG = np.meshgrid(xSeq, ySeq)

    plot.ion()
    plot.pcolor(xG, yG, zG_A, cmap = cmap)
    #plot.axis([0, n-1, n-1, 0])
    plot.axis([0, n, n, 0])
    plot.axes().set_aspect('equal')

    plot.title(plotTitle)

    raw_input("Hit Enter to dismiss the plot.")
    plot.close()

#==========================================================================================================
# MAINLINE
#==========================================================================================================
distanceThreshold = 16.0 # Any distance > distanceThreshold will be represented as white.

pdbIDchars = raw_input("Type in 4 character PDB ID or PDB file name: \n")
prot = chimera.openModels.open(pdbIDchars, type="PDB")[0]

chainIds = raw_input("Type in a single character or multiple characters separated by spaces to select one or more chains: \n")
chainIds_L = chainIds.split()

alphaCarb_L = []
mappedDist_D = {}
ix_D = {}

# Strategy: We use the Shell class to rapidly access neighbouring atoms that are within the
# distance threshold.  This means that we can avoid a full O(n**2) cost of evaluating all
# possible pairwise distances.
# The dictionary ix_D will give us a mapping between an atom and its index.  Atoms have an index
# that corresponds to the order in which they are found (usually this correpsonds in some simple
# way to the residue position number).
# The index is used in the plot function to give an atom a row and column index for the plot.

ix = 0
for r in prot.residues:
    if not r.id.chainId in chainIds_L: continue
    aca = r.findAtom("CA")
    if aca:
        alphaCarb_L.append(aca)
        ix_D[aca] = ix
        ix += 1

shell = Shell(alphaCarb_L, 0.0, distanceThreshold)

# Evaluate distances between the alpha carbon atoms and set them up in the mapped_Dist_D[]dictionary.
# The distances are dictionary values corresponding to keys that are pairs of residue positions.
# Distances are normalized before being passed to the generate2Dplot function.
# Consequently, all distances will map to a number be in the range [0.1, 1.0] and any distances
# over the contactThreshold will effectively map to 1.0.

for alpha in alphaCarb_L:
    mappedDist_D[tuple((ix_D[alpha], ix_D[alpha]))] = 0.0
    for nbr in shell.getAtomsInShell(alpha):
        dist = (numpy.linalg.norm(alpha.coord() - nbr.coord()))/distanceThreshold
        mappedDist_D[tuple((ix_D[alpha], ix_D[nbr]))] = dist
    
numAlphas = len(alphaCarb_L)

distMap2D_A = buildDistMapArray(numAlphas, mappedDist_D)

generate2Dplot(distMap2D_A, "Contact plot for " + pdbIDchars)















