import chimera, numpy
from numpy import array, linalg
from sets import Set

from StructBio.HBonds import HBondDicts
from StructBio.Utilities.shellNeighbors import Shell


# -----------------------------------------------------------------------------
# Build an atom information string when given the atom object.
def atomInfoString(atom):
    resType = atom.residue.type
    resPosn = atom.residue.id.position
    chainID = atom.residue.id.chainId
    atmName = atom.name
    retStr = '%s %d.%1s %-3s' % (resType, resPosn, chainID, atmName)
    return retStr.ljust(15,' ')

#==============================================================================
# AtomPair CLASS
#==============================================================================
# This is essentially a "struct" that retains the info for a single output line.
class AtomPair(object):
    def __init__(self, atomA, atomB, h_bonds):
        self.atomA = atomA
        self.atomB = atomB
        self.atomA_infoStr = atomInfoString(self.atomA)
        self.atomB_infoStr = atomInfoString(self.atomB)
        self.distance = linalg.norm(array(atomA.coord()) - array(atomB.coord()))
        # Codes for atom/atom interaction:
        #   B: Bonded (the two atoms are covalently bonded)
        #   G: Geminals (the two atoms are both bonded to a third atom)
        #   H: Hydrogen bonded (the two atoms have a hydrogen bond between them)
        #   N: None of the above.
        self.pairInteractionType = "N"
        if self.atomB in self.atomA.neighbors: self.pairInteractionType = "B"
        if len(Set(self.atomA.neighbors).intersection(Set(self.atomB.neighbors))) > 0:
            self.pairInteractionType = "G"
        if self.atomA in h_bonds.hBond_D.keys():
            if self.atomB in h_bonds.hBond_D[self.atomA]:
                self.pairInteractionType = "H"
        if self.atomB in h_bonds.hBond_D.keys():
            if self.atomA in h_bonds.hBond_D[self.atomB]:
                self.pairInteractionType = "H"

#==============================================================================
# GenerateDistanceData CLASS
#==============================================================================
class GenerateDistanceData(object):
    def __init__(self, protMol, threshold):
        self.protMol = protMol
        self.threshold = threshold
        self.h_Bonds = HBondDicts(protMol)
        self.atomPairs_L = []

        self.shellNbrs = Shell(self.protMol.atoms, 0.0, self.threshold)
        for centerAtom in self.protMol.atoms:
            for nbrAtom in self.shellNbrs.getAtomsInShell(centerAtom):
                self.atomPairs_L.append(AtomPair(centerAtom, nbrAtom, self.h_Bonds))

#==============================================================================
pdbIDchars = raw_input("Type in 4 character PDB ID or PDB file name: \n")
prot = chimera.openModels.open(pdbIDchars, type="PDB")[0]
threshold = raw_input("Type in floating point number for threshold: \n")

distData = GenerateDistanceData(prot, float(threshold))

threshStr = str(float(threshold)).replace('.', '_')
outFilePath = "NN_" + threshStr + "_" + pdbIDchars[0:4] + ".txt"
outFile = file(outFilePath, 'w')

for pair in distData.atomPairs_L:
    outFile.write('%s %s %6.3f %s\n' % (pair.atomA_infoStr,
                                        pair.atomB_infoStr,
                                        pair.distance,
                                        pair.pairInteractionType))

outFile.close()
                  





















