import chimera, numpy
from numpy import *

# -----------------------------------------------------------------------------
# For a set of 4 consecutive atoms, compute the dihedral angle.
# The input is a list of atoms.
# The angle is returned in degrees.
def dihedralAngle(atoms_L):
    u = atoms_L[1].coord() - atoms_L[0].coord()
    v = atoms_L[2].coord() - atoms_L[1].coord()
    w = atoms_L[3].coord() - atoms_L[2].coord()
    uCROSSv = tuple([u[1]*v[2]-u[2]*v[1], u[2]*v[0]-u[0]*v[2], u[0]*v[1]-u[1]*v[0]])
    vCROSSw = tuple([v[1]*w[2]-v[2]*w[1], v[2]*w[0]-v[0]*w[2], v[0]*w[1]-v[1]*w[0]])
    uCROSSvN = uCROSSv/linalg.norm(uCROSSv)
    vCROSSwN = vCROSSw/linalg.norm(vCROSSw)
    angle = numpy.sign(numpy.dot(uCROSSvN, w)) * math.acos(dot(uCROSSvN, vCROSSwN))
    return angle*180/math.pi

# -----------------------------------------------------------------------------
# Generate a "label" character string for a residue:
def resInfo(rr):
    return rr.type + " " + str(rr.id.position) + "." + rr.id.chainId

#==============================================================================
# Mainline:
#==============================================================================

pdbID = raw_input("Type in a PDB ID \n")  # Try 1BQY, 1TCR, 1A4Y, 1ZNJ
prot = chimera.openModels.open(pdbID, type = 'PDB')[0]

noBridgeResidues_L = []
bridges_D = {}
dihedrals_D = {}

for r in prot.residues:
    if r.type != "CYS": continue
    sulfurAtom = r.findAtom('SG')

    bridgeFound = False
    for sNbr in sulfurAtom.neighbors:
        if sNbr.name == 'SG':
            bridgeFound = True
            bridgePartner = sNbr

    if not bridgeFound:
        noBridgeResidues_L.append(r)
        continue

    partnerRes = bridgePartner.residue
    # If we get this far, we have a cysteine bridge.
    # To process a bridge only once, we demand that the residue
    # holding the other SG atom has a position not already processed:
    if r.id.chainId > partnerRes.id.chainId: continue
    if r.id.position > partnerRes.id.position: continue

    # Compute distance between alpha carbon atoms in the CYS residues:
    alphaDiff = array(r.findAtom('CA').coord()) - array(partnerRes.findAtom('CA').coord())
    bridges_D[(r, partnerRes)] = numpy.linalg.norm(alphaDiff)

    # Compute dihedral angle:
    dihedrals_D[(r, partnerRes)] = dihedralAngle([r.findAtom('CB'), r.findAtom('SG'),
                                                partnerRes.findAtom('SG'), partnerRes.findAtom('CB')])


# Report results:
if len(bridges_D) > 0:
    print "\nDisulfide bridges: \n"
    print "      Residue pairs          Distance between     Dihedral"
    print "                               alpha carbons       Angles\n"
    for k in bridges_D.keys():
        resLabel = resInfo(k[0])
        parLabel = resInfo(k[1])
        print resLabel, "\t", parLabel, " \t %7.3f" % bridges_D[k], "     \t %7.3f" % dihedrals_D[k]

if len(noBridgeResidues_L) > 0:
    print "\n\nThere is no disulfide bridge for each of the following: \n"
    for r in noBridgeResidues_L:
        print  resInfo(r)


