import chimera, numpy
from numpy import array, linalg, dot, arccos
from StructBio.LeastSquares.planes import LeastSquaresPlane

# We will use the term "ligand" atoms when referring to CMO atoms.
# NOTE: the least squares plane is drawn only for the first chain.

# The Kachalova et al. paper [KP99] references PDB ID = 1BZR
pdbIDchars = raw_input("Type in 4 character PDB ID or PDB file name: \n") 
prot = chimera.openModels.open(pdbIDchars, type="PDB")[0]

# Generate least squares plane object:
# While getting the nitrogen atoms that define the plane, we also collect other
# atoms needed for the angle calculations.
chainID = prot.sequences(asDict = True).keys()[0]
targetNames_L = ["NA", "NB", "NC", "ND"]
targetAtoms_L = []
for r in prot.residues:
    # Look for HEM "residue".
    if r.type == "HEM" and r.id.chainId == chainID:
        fe_inHEM = r.findAtom('FE')
        na_inHEM = r.findAtom('NA')
        nc_inHEM = r.findAtom('NC')
        for a in r.atoms:
            if a.name in targetNames_L:
                targetAtoms_L.append(a)

    # Look for carbon monoxide ligand:                
    if r.type == "CMO":
        c_inCMO = r.findAtom('C')
        o_inCMO = r.findAtom('O')

lsPlane = LeastSquaresPlane(targetAtoms_L)

# We have all the required atoms: fe_inHEM, nc_inHEM, etc.
# Compute the projections of their atomic coordinates onto the LS plane.
# NOTE: the projections will be relative to the PDB frame of reference.

fe_inHEMproj = lsPlane.projectionOnPlane(fe_inHEM)  # This will be our new origin.
na_inHEMproj = lsPlane.projectionOnPlane(na_inHEM)
nc_inHEMproj = lsPlane.projectionOnPlane(nc_inHEM)  # NA, NC define the x-axis direction.
c_inCMOproj = lsPlane.projectionOnPlane(c_inCMO)
o_inCMOproj = lsPlane.projectionOnPlane(o_inCMO)

# We now calculate the various angles defined in the paper by Kachalova et al. [KP99].

# Calculte the alpha angle:
#   This is 90 - the angle made by the CMO bond and its projection in the plane.
alphaAngleComp = lsPlane.angleToPlane(c_inCMO, o_inCMO)
print "Alpha angle for CMO within " + pdbIDchars[0:4] + " = %7.3f" % (90.0 - alphaAngleComp)

# Calculate the tau angle:
# This is 90 - the angle made by the line [fe_inHEMproj - c_inCMO.coord()) and its
# projection in the plane.
tauAngleComp = lsPlane.angleToPlane(fe_inHEM, c_inCMO)
print "Tau angle for C in CMO within " + pdbIDchars[0:4] + " = %7.3f" % (90.0 - tauAngleComp)

# Calculate the azimuthal phi angles:
x_axis = nc_inHEMproj - na_inHEMproj
x_axisUnit = x_axis/linalg.norm(x_axis)
# Phi_FeC:
fe_cDir = c_inCMOproj - fe_inHEMproj # Direction in the plane
fe_cUnit = fe_cDir/linalg.norm(fe_cDir)
print "Phi_FeC angle = %7.3f" % (numpy.arccos(dot(x_axisUnit, fe_cUnit))*180/numpy.pi)
# Phi_CO:
c_oDir = o_inCMOproj - c_inCMOproj   # Direction in the plane
c_oUnit = c_oDir/linalg.norm(c_oDir)
print "Phi_CO angle = %7.3f" % (numpy.arccos(dot(x_axisUnit, c_oUnit))*180/numpy.pi)





