import chimera, numpy as np
from chimera import runCommand
from StructBio.LeastSquares.inertialAxes import InertialAxes
from StructBio.Scenographics.ellipsoid import Ellipsoid
from StructBio.Scenographics.solids import Solids
from StructBio.StructureOverlap.overlapper import Overlapper

#===============================================================================
# GUI:
#===============================================================================
# GUI imports:
import Tkinter as tk, tkColorChooser, tkFileDialog
from chimera.baseDialog import ModelessDialog

#===============================================================================
# DomainManager class
#===============================================================================
class DomainManager(object):
    def __init__(self, domainDesignation, residues_L, parentFrame):

        self.domainDesignation = domainDesignation
        self.residues_L = residues_L
        self.ellipsoid = None

        #-------------------------------------------------------------------------
        # Domain FRAME
        eR_topText = tk.Label(parentFrame, text =  " Domain " + domainDesignation)
        eR_topText.grid(row = 0, column = 0, pady = 4, sticky = tk.W)

        # Compute inertial axes for this domain:
        atoms_L = []
        for r in residues_L: atoms_L.extend(r.atoms)
        self.axesObj = InertialAxes(atoms_L, massWeighted = True)

        #-------------------------------------------------------------------------
        # Ellipsoid Mesh subframe
        ellipsoidMesh_frame = tk.Frame(parentFrame, bd = 2, relief = tk.RIDGE)
        ellipsoidMesh_frame.grid(row = 1, column = 0, padx = 4, pady = 4, sticky = tk.W + tk.E)

        # Salutation
        eMText = tk.Label(ellipsoidMesh_frame, text =  " Ellipsoid Mesh: ")
        eMText.grid(row = 0, column = 0, columnspan = 2, pady = 4, sticky = tk.W)

        # Meridians spinbox
        eMMerText = tk.Label(ellipsoidMesh_frame, text =  " Meridians: ")
        eMMerText.grid(row = 1, column = 0, pady = 4, sticky = tk.W)
        self.meridianCount = tk.StringVar()
        tk.Spinbox(ellipsoidMesh_frame, width = 3, from_ = 4, to = 128, increment = 4, 
                   bg = "white", textvariable = self.meridianCount).grid(row = 1, column = 1)
        self.meridianCount.set("32")

        # Parallels spinbox
        eMParText = tk.Label(ellipsoidMesh_frame, text =  "       Parallels: ")
        eMParText.grid(row = 1, column = 2, pady = 4, sticky = tk.W)
        self.parallelCount = tk.StringVar()
        tk.Spinbox(ellipsoidMesh_frame, width = 3, from_ = 4, to = 128, increment = 4, 
                   bg = "white", textvariable = self.parallelCount).grid(row = 1, column = 3)
        self.parallelCount.set("8")

        # Mesh color
        eMColorText = tk.Label(ellipsoidMesh_frame, text =  "       Color: ")
        eMColorText.grid(row = 1, column = 4, pady = 4, sticky = tk.W)
        self.meshColor = (.0, .8, .0)
        self.meshColorBtn = tk.Button(ellipsoidMesh_frame, text = "", width = 5, height = 1, command = self.meshColor_btn_Func)
        self.meshColorBtn.configure(bg = colorDec2Hex(self.meshColor))
        self.meshColorBtn.grid(row = 1, column = 5, padx = 4, pady = 4, sticky = tk.W)

        # Mesh accept        
        self.meshAccept_btn = tk.Button(ellipsoidMesh_frame, text = "Accept", state = tk.NORMAL,
                                           command = self.meshAccept_btn_Func)
        self.meshAccept_btn.grid(row = 1, column = 6, columnspan = 2, padx = 8, pady = 4, sticky = tk.E)

 
        #-------------------------------------------------------------------------
        # Ellipsoid Surface subframe
        ellipsoidSurface_frame = tk.Frame(parentFrame, bd = 2, relief = tk.RIDGE)
        ellipsoidSurface_frame.grid(row = 2, column = 0, padx = 4, pady = 4, sticky = tk.W + tk.E)

        # Salutation
        eSText = tk.Label(ellipsoidSurface_frame, text =  " Ellipsoid Surface: ")
        eSText.grid(row = 0, column = 0, columnspan = 2, pady = 4, sticky = tk.W)

        # Surface color
        eSColorText = tk.Label(ellipsoidSurface_frame, text =  " Color: ")
        eSColorText.grid(row = 1, column = 0, pady = 4, sticky = tk.W)
        self.surfaceColor = (.3, .6, .9)
        self.surfaceColorBtn = tk.Button(ellipsoidSurface_frame, text = "", width = 5, height = 1, command = self.surfaceColor_btn_Func)
        self.surfaceColorBtn.configure(bg = colorDec2Hex(self.surfaceColor))
        self.surfaceColorBtn.grid(row = 1, column = 1, padx = 4, pady = 4, sticky = tk.W)
        
        # Surface transparency
        self.transparency_scaleVar = tk.DoubleVar()
        transparency_text = tk.Label(ellipsoidSurface_frame, text =  " Transparency:  ")
        transparency_text.grid(row = 1, column = 2, pady = 4, sticky = tk.E)
        self.transparency_scale = tk.Scale(ellipsoidSurface_frame, variable = self.transparency_scaleVar,
                                           orient = tk.HORIZONTAL, length = 100, from_ = 0.1, to = 1.0, resolution = 0.1,
                                           troughcolor = "white")
        self.transparency_scale.grid(row = 1, column = 3, columnspan = 3, sticky = tk.W)
        self.transparency_scale.set(0.2)
        
        # Surface accept
        self.surfaceAccept_btn = tk.Button(ellipsoidSurface_frame, text = "Accept", state = tk.NORMAL,
                                           command = self.surfaceAccept_btn_Func)
        self.surfaceAccept_btn.grid(row = 1, column = 6, padx = 8, pady = 4, sticky = tk.E)
        
        #-------------------------------------------------------------------------
        # Show_Hide subframe
        showHide_frame = tk.Frame(parentFrame, bd = 2, relief = tk.RIDGE)
        showHide_frame.grid(row = 3, column = 0, padx = 4, pady = 4, sticky = tk.W + tk.E)

        # Salutation
        sSText = tk.Label(showHide_frame, text =  " Show Hide Graphics: ")
        sSText.grid(row = 0, column = 0, columnspan = 2, pady = 4, sticky = tk.W)
        
        # Mesh show button:
        self.showMeshChecked = tk.BooleanVar()
        cbM = tk.Checkbutton(showHide_frame, text = " Mesh     ", variable = self.showMeshChecked,
                             command = self.showMeshChecked_func, indicatoron = 1)
        cbM.grid(row = 1, column = 0, pady = 6, sticky = tk.E)

         
        # Surface show button:
        self.showSurfaceChecked = tk.BooleanVar()
        cbS = tk.Checkbutton(showHide_frame, text = " Surface     ", variable = self.showSurfaceChecked,
                             command = self.showSurfaceChecked_func, indicatoron = 1)
        cbS.grid(row = 1, column = 1, pady = 6, sticky = tk.E)

         
        # Axes show button:
        self.showAxesChecked = tk.BooleanVar()
        cbA = tk.Checkbutton(showHide_frame, text = " Axes     ", variable = self.showAxesChecked,
                             command = self.showAxesChecked_func, indicatoron = 1)
        cbA.grid(row = 1, column = 2, pady = 6, sticky = tk.E)

        
        #-------------------------------------------------------------------------
        # GUI is established.  Put graphics into the scene (but hidden)
        self.setUpAxes()
        self.meshAccept_btn_Func()  # Note: we need the mesh before a surface can be generated.
        self.ellipsoid.displaySurface(self.surfaceColor, transparency = self.transparency_scale.get())

        self.axes.hide()
        self.ellipsoid.hideMesh()
        self.ellipsoid.hideSurface()
        
    #-------------------------------------------------------------------------
    # Callback functions for DomainManager widgets
    #-------------------------------------------------------------------------
    def meshColor_btn_Func(self):
        (rgb, hx) = tkColorChooser.askcolor(initialcolor = colorDec2Hex(self.meshColor), title = "Choose a color")
        self.meshColor = (rgb[0]/255., rgb[1]/255., rgb[2]/255.)
        self.meshColorBtn.configure(bg = hx)
        self.meshAccept_btn_Func()

    #-------------------------------------------------------------------------
    def meshAccept_btn_Func(self):
        if self.ellipsoid:
            self.ellipsoid.removeMesh()
            self.ellipsoid.removeSurface()
            
        # Now redeploy the ellipsoid:
        (length, width, height) = self.getEllipsoidDimensions()
        self.ellipsoid = Ellipsoid(length, width, height, self.axesObj.centroid, self.axesObj.inertialAxes_L,
                                   int(self.meridianCount.get()), int(self.parallelCount.get()),
                                   ellipsoidName = self.domainDesignation + " ellipsoid")
        self.ellipsoid.displaySurface(self.surfaceColor, transparency = self.transparency_scale.get())
        self.ellipsoid.displayMesh(self.meshColor)
        # Be consistent with checkbuttons:
        self.showMeshChecked_func()
        self.showSurfaceChecked_func()

    #-------------------------------------------------------------------------
    def surfaceColor_btn_Func(self):
        (rgb, hx) = tkColorChooser.askcolor(initialcolor = colorDec2Hex(self.surfaceColor), title = "Choose a color")
        self.surfaceColor = (rgb[0]/255., rgb[1]/255., rgb[2]/255.)
        self.surfaceColorBtn.configure(bg = hx)
        self.surfaceAccept_btn_Func()

    #-------------------------------------------------------------------------
    def surfaceAccept_btn_Func(self):
        self.ellipsoid.removeSurface()                                    
        self.ellipsoid.displaySurface(self.surfaceColor, transparency = self.transparency_scale.get())
        # Be consistent with checkbuttons:
        self.showSurfaceChecked_func()
        
    #-------------------------------------------------------------------------
    def showMeshChecked_func(self):
        if self.showMeshChecked.get(): self.ellipsoid.showMesh()
        else: self.ellipsoid.hideMesh()
        
    #-------------------------------------------------------------------------
    def showSurfaceChecked_func(self):
        if self.showSurfaceChecked.get(): self.ellipsoid.showSurface()
        else: self.ellipsoid.hideSurface()

    #-------------------------------------------------------------------------
    def showAxesChecked_func(self):
        if self.showAxesChecked.get(): self.axes.show()
        else: self.axes.hide()
                                    
        
    #-------------------------------------------------------------------------
    # Utility functions for DomainManager widgets
    #-------------------------------------------------------------------------
    def setUpAxes(self):
        colors_L = [(1.,0.,0.), (0.,1.,0.), (0.,0.,1.)]
        self.axes = Solids(modelName = self.domainDesignation + " Axes")
        for i in range(3):
            ends = self.axesObj.min_maxProjection(i)
            self.axes.addSpindle(self.axesObj.inertialAxes_L[i] * ends[0] + self.axesObj.centroid,
                          self.axesObj.inertialAxes_L[i] * ends[1] + self.axesObj.centroid,
                          .3, colors_L[i])
        self.axes.display()

    #-------------------------------------------------------------------------

    def getEllipsoidDimensions(self):
        sq5 = np.sqrt(5.0)  # Scaling factor used by Browner et al.
        length = sq5 * self.axesObj.shape_weightedAxisLength(0)
        width = sq5 * self.axesObj.shape_weightedAxisLength(1)
        height = sq5 * self.axesObj.shape_weightedAxisLength(2)
        return (length, width, height)

#===============================================================================
# ProtManager class
#===============================================================================
class ProtManager(object):
    def __init__(self, parent, protType, parentFrame):
        self.parent = parent
        self.protType = protType
        self.parentFrame = parentFrame
        self.otherManager = None
        self.prot = None
        
        #-------------------------------------------------------------------------
        # FETCH PDB data FRAME
        fetchPDBdata_frame = tk.Frame(self.parentFrame, bd = 2, relief = tk.RIDGE) #, bg = bgColor)
        fetchPDBdata_frame.grid(row = 0, column = 0, columnspan = 2, padx = 4, pady = 4, sticky = tk.W + tk.E)
         
        fetch_topText = tk.Label(fetchPDBdata_frame, text =  " Fetch " + self.protType + " PDB File ")
        fetch_topText.grid(row = 0, column = 0, columnspan = 3, pady = 4, sticky = tk.W)

        # Gather info specifying protein data:
        tk.Label(fetchPDBdata_frame, text =  " " + self.protType + " PDB ID: ").grid(row = 1, column = 0, pady = 4, sticky = tk.W)
        self.pdb_ID_entryVar = tk.StringVar()
        eP = tk.Entry(fetchPDBdata_frame, textvariable = self.pdb_ID_entryVar, width = 7, bg = "white")
        eP.grid(row = 1, column = 1, padx = 4, pady = 4, sticky = tk.E)

        tk.Label(fetchPDBdata_frame, text =  "  Chain ID: ").grid(row = 1, column = 2, pady = 4, sticky = tk.W)
        self.chain_ID_entryVar = tk.StringVar()
        eC = tk.Entry(fetchPDBdata_frame, textvariable = self.chain_ID_entryVar, width = 7, bg = "white")
        eC.grid(row = 1, column = 3, padx = 4, pady = 4, sticky = tk.E)

        tk.Label(fetchPDBdata_frame, text =  "  Domain D1: ").grid(row = 1, column = 4, pady = 4, sticky = tk.W)
        self.d1_entryVar = tk.StringVar()
        eD1 = tk.Entry(fetchPDBdata_frame, textvariable = self.d1_entryVar, width = 16, bg = "white")
        eD1.grid(row = 1, column = 5, padx = 4, pady = 4, sticky = tk.E)

        tk.Label(fetchPDBdata_frame, text =  "  Domain D2: ").grid(row = 1, column = 6, pady = 4, sticky = tk.W)
        self.d2_entryVar = tk.StringVar()
        eD2 = tk.Entry(fetchPDBdata_frame, textvariable = self.d2_entryVar, width = 16, bg = "white")
        eD2.grid(row = 1, column = 7, padx = 4, pady = 4, sticky = tk.E)

        self.btnFetch = tk.Button(fetchPDBdata_frame, text = " Fetch ", command = self.btnFetch_Func)
        self.btnFetch.grid(row = 1, column = 8, padx = 4, pady = 4)

        if self.protType == "Holo": self.btnFetch.config(state = tk.DISABLED)

        
    #-------------------------------------------------------------------------
    # Callback functions for ProtManager widgets
    #-------------------------------------------------------------------------
    def btnFetch_Func(self):
        # Fetch the protein and eliminate anything outside of the specified chain.
        self.prot = chimera.openModels.open(self.pdb_ID_entryVar.get(), type="PDB")[0]
        commandLine = "sel #" + str(self.prot.id) + ";"
        commandLine += " ~sel #" + str(self.prot.id) + ":." + self.chain_ID_entryVar.get() + ";"
        commandLine += " delete sel; ~sel"
        runCommand(commandLine)

        # Declare both frames for the domains:
        domainD1_frame = tk.Frame(self.parentFrame, bd = 2, relief = tk.RIDGE)
        domainD1_frame.grid(row = 1, column = 0, padx = 4, pady = 4, sticky = tk.W + tk.E)
        domainD2_frame = tk.Frame(self.parentFrame, bd = 2, relief = tk.RIDGE)
        domainD2_frame.grid(row = 1, column = 1, padx = 4, pady = 4, sticky = tk.W + tk.E)
        
        self.domainD1residues_L = self.getResiduesInDomain(self.d1_entryVar.get())
        self.domainD2residues_L = self.getResiduesInDomain(self.d2_entryVar.get())

        if self.prot and self.protType == "Apo":
            self.d1_DomainManager = DomainManager("D1", self.domainD1residues_L, domainD1_frame)
            self.d2_DomainManager = DomainManager("D2", self.domainD2residues_L, domainD2_frame)            
        
        # If this is the holo protein manager, then immediately move to overlap:
        if self.protType == "Holo" and self.otherManager.prot:
            # If we get this far then both proteins have been fetched and we can do the overlap.
            self.parent.overlapHoloOverApo()
            # Now that we have overlap (and hence, new coordinates for the Holo), build the domain managers:
            self.d1_DomainManager = DomainManager("D1", self.domainD1residues_L, domainD1_frame)
            self.d2_DomainManager = DomainManager("D2", self.domainD2residues_L, domainD2_frame)
            # Fill in results frame at bottom of GUI:
            self.parent.showRMSDforD2movement()
            self.parent.showBendforD2movement()

        # Disable the fetch button and enable the holo fetch button.
        self.btnFetch.config(state = tk.DISABLED)
        if self.prot and self.protType == "Apo":
            self.otherManager.btnFetch.config(state = tk.NORMAL)

        
    #-------------------------------------------------------------------------
    # Utility functions for ProtManager widgets
    #-------------------------------------------------------------------------
    def getResiduesInDomain(self, domainDescriptor):
        # The domainDescriptor is a character string with the format start1-end1,start2-end2, etc.
        # The domainDescriptor is a comma separated set of ranges.
        # Each range is a minus sign separated pair of position numbers specify a range of  
        # consecutive residues contributing to the domain.
        
        resPositions_L = []
        ranges = [range(int(x), int(y) + 1) for [x,y] in [x.split("-") for x in domainDescriptor.split(",")]]
        for x in ranges: resPositions_L.extend(x)
        residues_L = []
        for r in self.prot.residues:
            if r.id.position in resPositions_L: residues_L.append(r)
        return residues_L

#===============================================================================
class ApoHoloMotionGUI(ModelessDialog):
    buttons = ("Close")
    help = False    # Leave this out to get a Help button.
    title = "Domain Motion"

    def fillInUI(self, parent):
        #-------------------------------------------------------------------------
        # APO super frame:
        apo_Super_frame = tk.Frame(parent, bd = 2, relief = tk.RIDGE, bg = "tan")
        apo_Super_frame.grid(row = 0, column = 0, padx = 4, pady = 4, sticky = tk.W + tk.E)
        self.apo_ProtManager = ProtManager(self, "Apo", apo_Super_frame) # Note: prot manager knows its parent.
        # Set up example:
        self.apo_ProtManager.pdb_ID_entryVar.set("1EX6")
        self.apo_ProtManager.chain_ID_entryVar.set("B")
        self.apo_ProtManager.d1_entryVar.set("201-232,284-386")
        self.apo_ProtManager.d2_entryVar.set("233-283")

        #-------------------------------------------------------------------------
        # HOLO super frame:
        holoSuper_frame = tk.Frame(parent, bd = 2, relief = tk.RIDGE, bg = "gold")
        holoSuper_frame.grid(row = 1, column = 0, padx = 4, pady = 4, sticky = tk.W + tk.E)
        self.holoProtManager = ProtManager(self, "Holo", holoSuper_frame)        
        # Set up example:
        self.holoProtManager.pdb_ID_entryVar.set("1EX7")
        self.holoProtManager.chain_ID_entryVar.set("A")
        self.holoProtManager.d1_entryVar.set("1-32,84-186")
        self.holoProtManager.d2_entryVar.set("33-83")

        # Enable manager communication:
        self.apo_ProtManager.otherManager = self.holoProtManager
        self.holoProtManager.otherManager = self.apo_ProtManager

        #-------------------------------------------------------------------------
        # Results super frame:
        resultsSuper_frame = tk.Frame(parent, bd = 2, relief = tk.RIDGE)
        resultsSuper_frame.grid(row = 2, column = 0, padx = 4, pady = 4, sticky = tk.W + tk.E)

        # Salutation
        rSText = tk.Label(resultsSuper_frame, text =  " Domain Movement Measurements ")
        rSText.grid(row = 0, column = 0, columnspan = 3, pady = 4, sticky = tk.W)

        # RMSD:
        rRMSDText = tk.Label(resultsSuper_frame, text =  " RMSD: ")
        rRMSDText.grid(row = 1, column = 0, pady = 4, sticky = tk.W)
        self.rRMSDTextResult = tk.Label(resultsSuper_frame, bg = "white")
        self.rRMSDTextResult.grid(row = 1, column = 1, pady = 4, sticky = tk.W)

        # D2 Bend:
        rBendText = tk.Label(resultsSuper_frame, text =  "        D2 Bend: ")
        rBendText.grid(row = 1, column = 2, pady = 4, sticky = tk.W)
        self.rBendTextResult = tk.Label(resultsSuper_frame, bg = "white")
        self.rBendTextResult.grid(row = 1, column = 3, pady = 4, sticky = tk.W)


    #-------------------------------------------------------------------------
    def domainAtomPairsList(self, resH_L, resA_L):
        # This function assumes that resH_L and resA_L have the same length and hold
        # the residues from the same domain: resH_L from the holo conformation and
        # resA_L from the apo conformation.  The return value is a list of alpha carbon
        # pairs for that domain.
        retList = []
        for ix in range(len(resH_L)):
            retList.append((resH_L[ix].findAtom("CA"), resA_L[ix].findAtom("CA")))
        return retList

    #-------------------------------------------------------------------------
    def overlapHoloOverApo(self):
        overlap_L = self.domainAtomPairsList(self.holoProtManager.domainD1residues_L,
                                             self.apo_ProtManager.domainD1residues_L)
        
        # Put in the next statement if an overlap is to include domain D2:
        overlap_L.extend(self.domainAtomPairsList(self.holoProtManager.domainD2residues_L,
                                                self.apo_ProtManager.domainD2residues_L))
        
        overlapper = Overlapper(self.holoProtManager.prot)
        overlapper.moveToOverlap(overlap_L, massWeighted = True)
        runCommand("focus")

    #-------------------------------------------------------------------------
    def showRMSDforD2movement(self):
        # First build list of corresponding atom pairs for domain D2:
        atomPairs_L = self.domainAtomPairsList(self.holoProtManager.domainD2residues_L,
                                               self.apo_ProtManager.domainD2residues_L)
        
        atomPairs_L.extend(self.domainAtomPairsList(self.holoProtManager.domainD1residues_L,
                                               self.apo_ProtManager.domainD1residues_L))
        
        sumSquaresDeviations = 0.0
        sumWeights = 0.0
        for (a,b) in atomPairs_L:
            # Print to see matching atom in pairs.
#            print np.linalg.norm(a.coord() - b.coord()), a.residue.id.position, b.residue.id.position, a.name
            sumSquaresDeviations += a.element.mass * np.linalg.norm(a.coord() - b.coord())**2
            sumWeights += a.element.mass
        self.rRMSDTextResult.configure(text = " %6.3f " % (np.sqrt(sumSquaresDeviations/sumWeights)))

    #-------------------------------------------------------------------------
    def showBendforD2movement(self):
        # Drill down to get the primncipal axes:
        resultString = ""
        for i in range(3):
            apo_D2Axis = self.apo_ProtManager.d2_DomainManager.axesObj.inertialAxes_L[i]
            holoD2Axis = self.holoProtManager.d2_DomainManager.axesObj.inertialAxes_L[i]
            dotProd = abs(np.dot(apo_D2Axis, holoD2Axis)) # Note: assuming angle between 0 and pi/2.
            resultString += "  %8.3f " % (np.arccos(dotProd)*180./np.pi)
        self.rBendTextResult.configure(text = resultString)

        
#===============================================================================
def colorDec2Hex(c_T):
    return "#%02x%02x%02x" % (int(c_T[0]*255), int(c_T[1]*255), int(c_T[2]*255))





#==============================================================================
# Mainline
#==============================================================================

runCommand("set bg_color white")


apo_holoMotionGUI = ApoHoloMotionGUI()





