import chimera, numpy, sets, Rotamers      #, operator
from chimera import runCommand
#from chimera.molEdit import addAtom, addBond
from numpy import array, linalg
from sets import Set
from Rotamers import getRotamers, useRotamer, NoResidueRotamersError
#from operator import itemgetter
#from MMMD import base

from StructBio.Utilities.shellNeighbors import Shell
from StructBio.EnergyFunctions.LinearEnergyFunctions import atomPairEnergy

#==============================================================================

def rotamerPairEnergy(aRes, bRes):
    # This function calculates the energy between two rotamers
    # taking into account the side chain atoms only
    # Backbone atoms NOT included.
    if aRes.type == 'GLY' or bRes.type == 'GLY': return 0.0
    
    energy = 0.0
    for a in aRes.atoms:
        if a.name in ('N', 'CA', 'C', 'O', 'HA', 'H',
                      'HA2', 'HA3', 'H1', 'H2', 'H3'): continue
        for b in bRes.atoms:
            if b.name in ('N', 'CA', 'C', 'O', 'HA', 'H',
                          'HA2', 'HA3', 'H1', 'H2', 'H3'): continue
            energy += atomPairEnergy(a, b)
    return energy

#-------------------------------------------------------------------------------------------
#
def rotamerSelfEnergy(aRes, statisticalIntrinsicEnergy, nearbyRes_L, fixedRes_L):
    # This calculates the sum of all atom pair energies such that one atom of a pair is in aRes 
    # (side chain only) and the other atom of a pair is a backbone atom in a residue of the nearbyRes_L
    # OR the other atom is in a sidechain of a residue in the fixedRes_L.
    # It is to be expected that fixedRes_L is a subset of the nearbyRes_L
    # Typically, the fixed residues will be nearby and will be GLY, ALA or CYS 
    # (cysteine considered a fixed residue if it is involved with a disulfide bridge to another CYS residue).
    
    energy = statisticalIntrinsicEnergy # This is the energy due to interactions between atoms within the sidechain.
    for a in aRes.atoms:
        if a.name in ('N', 'CA', 'C', 'O', 'HA', 'H', 'HA2', 'HA3', 'H1', 'H2', 'H3'): continue # Skip backbone atoms of aRes.

        # Other atom is in the backbone of residues in nearbyRes_L:
        for nrbyR in nearbyRes_L:
            if abs(nrbyR.id.position - aRes.id.position) <= 1: continue # Skip nearby backbone (see pg. 32 of Kingsford thesis).
            for b in nrbyR.atoms:
                if b.name in ('N', 'CA', 'C', 'O', 'HA', 'H', 'HA2', 'HA3', 'H1', 'H2', 'H3'):
                    energy += atomPairEnergy(a, b)
                    
        # Other atom is in the sidechain of residues in fixedRes_L:
        for fixedR in fixedRes_L:
            for b in fixedR.atoms:
                if b.name in ('N', 'CA', 'C', 'O', 'HA', 'H', 'HA2', 'HA3', 'H1', 'H2', 'H3'): continue
                energy += atomPairEnergy(a, b)

    return energy

#==============================================================================
# SingleRotamer CLASS
#==============================================================================

class SingleRotamer(object):
    # Tracks all the information related to a single rotamer at a particular residue site.
    def __init__(self, ix, siteResidue, statistical_IntrinsicEnergy, originalChi1, originalChi2):
        self.rotamerIndex = ix  # Needed in some applications as a rotamer identifier.
        self.siteResidue = siteResidue
        self.rotamer = getRotamers(self.siteResidue)[1][self.rotamerIndex]
        self.statistical_IntrinsicEnergy = statistical_IntrinsicEnergy

        # The next variable is true if the rotamer angle chi1 is within 40 degrees of the original chi1.
        self.within40Chi1 = None
        # The next variable is true if ch2 is both within 40 degress of the original AND
        # within40chi1 is True.  It stays at None if the rotamer does not have a chi2 angle.
        self.within40Chi2 = None

        useRotamer(self.siteResidue, [self.rotamer])
        if originalChi1:
            chi1Diff = abs(originalChi1 - self.siteResidue.chi1)
            if chi1Diff > 180: chi1Diff = 360 - chi1Diff
            self.within40Chi1 = chi1Diff <= 40
        if originalChi2:
            chi2Diff = abs(originalChi2 - self.siteResidue.chi2)
            if chi2Diff > 180: chi2Diff = 360 - chi2Diff
            self.within40Chi2 = originalChi1 and chi2Diff <= 40
        # The next statement assumes that the useRotamer statement has set up the residue site
        # with this rotamer.
        self.selfEnergy = rotamerSelfEnergy(self.siteResidue, self.statistical_IntrinsicEnergy,
                                            nbrsMap_D_G[self.siteResidue], fixedRes_L_G)

        # Printing to track progress of rotamer generation:
        # Headers: "Type:   Pos:   Number:  Energy:"
        print " %3s   %4d      %2d   %9.4f" % (self.siteResidue.type, self.siteResidue.id.position,
                                               self.rotamerIndex, self.selfEnergy)

    def updateRotPairEnergyDictionary(self):
        global allRotamers_G, edgeCount_G, rotamerPairEnergy_D_G
        # This can only be called after all the Rotamer objects have been instantiated.
        # Compute the energy interactions between this rotamer and all the rotamers that
        # are at positions corresponding to nearby residues; these values to be stored in
        # the global dictionary rotamerPairEnergy_D_G.
        # Key is a tuple: (residue position, rotamer number, residue position, rotamer number).
        # Value is the corresponding pair energy.
        #
        # Skip if this ResRots object corresponds to a glycine or an alanine
        #   (since there are no rotamers for these residues).
        if self.siteResidue.type == "ALA" or self.siteResidue.type == "GLY": return
        
        useRotamer(self.siteResidue, [self.rotamer])
        for resNbr in nbrsMap_D_G[self.siteResidue]:
            if not resNbr in residuesWithRotamers_L_G: continue
            # Do calculation only for neighbouring residues that are further along in position:
            if resNbr.id.position <= self.siteResidue.id.position: continue
            
            for nbrRotamerIx in allRotamers_G.siteRotamers_D[resNbr].singleRotamers_D.keys():
                useRotamer(resNbr, [allRotamers_G.siteRotamers_D[resNbr].singleRotamers_D[nbrRotamerIx].rotamer])
                energy = rotamerPairEnergy(self.siteResidue, resNbr)
                rotamerPairEnergy_D_G[(self.siteResidue.id.position, self.rotamerIndex,
                                       resNbr.id.position, nbrRotamerIx)] = energy

                if energy != 0.0:
                    edgeCount_G += 1
                    
#==============================================================================
# SiteRotamers CLASS
#==============================================================================

class SiteRotamers(object):
    def __init__(self, res):
        global collisionThreshold_G
        self.res = res
        originalChi1 = res.chi1
        originalChi2 = res.chi2

        # Retain original chi angles for future reference:
        self.chi1 = res.chi1
        self.chi2 = res.chi2
        
        # Rotamers may be deleted due to Dead End Elimination so we track
        # the current rotamers for a site by using a dictionary:
           # Key is the original rotamer list index.
           # Value is a SingleRotamer object.
        self.singleRotamers_D = {}
        
        if res.type == "GLY" or res.type == "ALA":
            self.rotCount = 0
            return
        
        rot_L = getRotamers(res)[1]
        self.rotCount = len(rot_L)

        p0 = rot_L[0].rotamerProb

        # Generate all the singleRotamer objects and set them up in the singleRotamers_D.
        for i in range(len(rot_L)):
            probThisRot = rot_L[i].rotamerProb
            if probThisRot > 0.0000000000001:
                statSelf_Energy = 10.0 * numpy.log(p0/probThisRot)  # See Kingsford's thesis, pg. 33
                singleRot = SingleRotamer(i, res, statSelf_Energy, originalChi1, originalChi2)
                if singleRot.selfEnergy - singleRot.statistical_IntrinsicEnergy < collisionThreshold_G:
                    self.singleRotamers_D[i] = singleRot
                else:
                    print res.type, res.id.position, "Deleting rotamer ", singleRot.rotamerIndex, " with intrinsic energy: ", singleRot.selfEnergy
                    del singleRot

    # -------------------------------------------------------------------------------------------
    # This function is needed by doGoldsteinDEE
    def __getMinDiffEnergyValues(self, rot_i, rot_j, resNbr):
        # Function to compute the min[] term within the sum of the Goldstein criterion 
        # (see Equation 7.3).
        # Input:
        #       rot_i:  index for rotamer i at this residue site
        #       rot_j:  index for rotamer j at this residue site
        #       resNbr: selects residue neighbor (designated by index k in Eq. 7.3)

        # First collect the energy difference values in a list:
        # Don't forget: the rotamerPairEnergy_D_G dictionary has a tuple key with residue positions
        # in increasing order.
        global allRotamers_G
        
        eDiffVals_L = []
        for rotXix in allRotamers_G.siteRotamers_D[resNbr].singleRotamers_D.keys():
            if resNbr.id.position < self.res.id.position:
                eDiffVals_L.append(rotamerPairEnergy_D_G[(resNbr.id.position, rotXix, self.res.id.position, rot_i)] -
                                   rotamerPairEnergy_D_G[(resNbr.id.position, rotXix, self.res.id.position, rot_j)])
            else:
                eDiffVals_L.append(rotamerPairEnergy_D_G[(self.res.id.position, rot_i, resNbr.id.position, rotXix)] -
                                   rotamerPairEnergy_D_G[(self.res.id.position, rot_j, resNbr.id.position, rotXix)])
        return min(eDiffVals_L)

    # -------------------------------------------------------------------------------------------
    # Function to calculate the LHS of the inequality (see Eq. 7.3).    
    def goldsteinCriterionValue(self, rot_i, rot_j):        
        # Input:
        #       rot_i:  index for rotamer i at this residue site
        #       rot_j:  index for rotamer j at this residue site
        # If the return value is positive then we can delete the rotamer indexed by rot_i.
        sumMinDiffEnergies = 0.0
        for rNbr in nbrsMap_D_G[self.res]:
            if rNbr == self.res: continue
            if not rNbr in residuesWithRotamers_L_G: continue
            sumMinDiffEnergies += self.__getMinDiffEnergyValues(rot_i, rot_j, rNbr)
        return self.singleRotamers_D[rot_i].selfEnergy - self.singleRotamers_D[rot_j].selfEnergy + sumMinDiffEnergies
        
    # -------------------------------------------------------------------------------------------
    # Do dead end elimination for this residue position following the DEE strategy described by Eq. 7.3
    # Instead of comparing every rotamer at this site with every other rotamer, we pick the
    # rotamer with the lowest intrinsic energy and test to see if any other rotamer at this site
    # can be eliminated.
    def doGoldsteinDEEcomparingWithLowEnergyRotomer(self):
        # Find rotamer with lowest intrinsic energy:
        lowestEvalue = float("inf")
        for rotIx in self.singleRotamers_D.keys():
            if self.singleRotamers_D[rotIx].selfEnergy < lowestEvalue:
                lowestEvalue = self.singleRotamers_D[rotIx].selfEnergy
                lowErotIx = rotIx
                
        for rotIx in self.singleRotamers_D.keys():
            if rotIx == lowErotIx: continue
            if self.goldsteinCriterionValue(rotIx, lowErotIx) > 0.0:
                # Header title: "Pos:   Number:"
                print "%4d      %2d" % (self.res.id.position, rotIx)
                self.singleRotamers_D.pop(rotIx)
                
    # -------------------------------------------------------------------------------------------
    # Do dead end elimination for this residue position following the DEE strategy described by Eq. 7.3
    # After applying doGoldsteinDEEcomparingWithLowEnergyRotomer(self) we run this
    # algorithm which compares each rotamer with all sibling rotamers.
    def doGoldsteinDEE_AllRotomersVsAllRotamers(self):
        rotamersDeleted = False
        victims_S = Set()
        for rot_i in self.singleRotamers_D.keys():
            for rot_j in self.singleRotamers_D.keys():
                if rot_i == rot_j: continue
                if self.goldsteinCriterionValue(rot_i, rot_j) > 0.0:
                    victims_S.add(rot_i)
                
        for rotIx in victims_S:
            # Header title: "Pos:   Number:"
            print "%4d      %2d" % (self.res.id.position, rotIx)
            self.singleRotamers_D.pop(rotIx)

        if len(victims_S) > 0: rotamersDeleted = True
        return rotamersDeleted
                
#==============================================================================
# AllRotamers CLASS
#==============================================================================

class AllRotamers(object):
    def __init__(self, res_L):
        self.res_L = res_L
        
        # Dictionary of SiteRotamers object:
           # Key is a residue.
           # Value is a SiteRotamers object.
        self.siteRotamers_D = {}

        for r in res_L:
            self.siteRotamers_D[r] = SiteRotamers(r)
            
    # -------------------------------------------------------------------------------------------
    # 
    def doEnergyCalculations(self):
        for r in self.res_L:
            for rotIx in self.siteRotamers_D[r].singleRotamers_D.keys():
                # Headers: "Type:   Pos:   Number:"
                print " %3s   %4d      %2d" % (r.type, r.id.position, rotIx)
                self.siteRotamers_D[r].singleRotamers_D[rotIx].updateRotPairEnergyDictionary()
                                
    # -------------------------------------------------------------------------------------------
    # 
    def doGoldsteinLowEnergyRotamer(self):
        for sR in self.siteRotamers_D.keys():
            self.siteRotamers_D[sR].doGoldsteinDEEcomparingWithLowEnergyRotomer()
                        
    # -------------------------------------------------------------------------------------------
    # 
    def doGoldsteinAllRotamersVsAllRotamers(self):
        rotamers_Deleted = True
        while rotamers_Deleted:
            for sR in self.siteRotamers_D.keys():
                rotamers_Deleted = self.siteRotamers_D[sR].doGoldsteinDEE_AllRotomersVsAllRotamers()
            
#============================================================================================
# MAINLINE
#============================================================================================

# Files previously used:
# "1TIM" "1C9O" "1BRF" "1AAC" "1AHO" "1B9O" "1C5E" "1CC7" "1CEX" "1CKU" "1CTJ" "1CZ9" "1CZP"
# "1CRN" "1D4T" "1IGD" "1MFM" "1PLC" "1QJ4" "1QTN" "1QU9" "1RCF" "1VFY" "2PTH" "4RXN" "5P21"
targetPath = "1C90"

outputFile = file("Results/" + targetPath + "data.txt", 'w')
outFormat = '%5d  %5d  %5d  %5d  %5d  %8.4f\n'

# -------------------------------------------------------------------------------------------
# Global variables:
# The fixed residues either have no rotamers or have a conformation that is not to be replaced
# by a rotamer (eg. cysteine with a bridge).
typesOfFixedResidues_G = ('GLY', 'ALA')

# If rotamer selfEnergy - rotamer statistical_IntrinsicEnergy >= collisionThreshold_G
# then the rotamer is colliding with fixed atoms and can be deleted.
collisionThreshold_G = 10.0

# Dictionary of rotamer pair energies.
#   Key is a tuple: (residue position, rotamer number, residue position, rotamer number).
#   Value is the corresponding pair energy.
rotamerPairEnergy_D_G = {}
protein_G = chimera.openModels.open(targetPath, type="PDB")[0]
# -------------------------------------------------------------------------------------------

threshold_G = 22.0 # Two residues are considered to be neigbors if their beta carbons have separation <= threshold_G
edgeCount_G = 0 # Used in printed output

runCommand('set bg_color white')
runCommand('~ribbon')
runCommand('represent stick')
runCommand('display')

# Build global dictionary that maps a residue (capable of having rotamers) to a list of neighbouring residues.
nbrsMap_D_G = {}
betaAtoms_L = []
for r in protein_G.residues:
    if r.type in ('GLY', 'ALA'): continue # Ignore if residue has no rotamers.
    if r.findAtom('CB'):
        betaAtoms_L.append(r.findAtom('CB'))
shell_G = Shell(betaAtoms_L, 0.0, threshold_G) # We can now use shell_G.getAtomsInShell(centerAtom) to get atom neighbors
for r in protein_G.residues:
    if r.type in ('GLY', 'ALA'): continue
    nbrs_L = []
    cBForThisRes = r.findAtom('CB')
    if cBForThisRes: 
        for cb in shell_G.getAtomsInShell(cBForThisRes):
            nbrs_L.append(cb.residue)
        print "\nResidue ", cBForThisRes.residue.type, " at position ", r.id.position, "has nearby residues:"
        outStr = ""
        for rs in nbrs_L:
            outStr += rs.type + repr(rs.id.position) + ", "
        print outStr[0:-2]
    nbrsMap_D_G[r] = nbrs_L

print "\nBuilding Rotamer Objects:"
print " Res    Res    Rotamer   Self"
print "Type:   Pos:   Number:  Energy:"

fixedRes_L_G = []
residuesWithRotamers_L_G = []
for r in protein_G.residues:
    if not r.findAtom('CA'): continue  # We only handle standard residues.
    if r.type in typesOfFixedResidues_G:
        fixedRes_L_G.append(r)
    else:
        residuesWithRotamers_L_G.append(r)
allRotamers_G = AllRotamers(residuesWithRotamers_L_G)

print "\nDoing Energy Calculations for Rotamers:"
print " Res    Res    Rotamer"
print "Type:   Pos:   Number:"
allRotamers_G.doEnergyCalculations()

# Record rotamer statistics prior to DEE:
outputSurvivalFile = file("Results/" + targetPath + "survivorCount.txt", 'w')
outSurvivalFormat = '%5d %4s %5d\n'
outputSurvivalFile.write("Prior to DEE:\n")
totalRotCount = 0
for res in allRotamers_G.siteRotamers_D.keys():
    rotamer_Count = len(allRotamers_G.siteRotamers_D[res].singleRotamers_D.keys())
    totalRotCount += rotamer_Count
    line_Out = outSurvivalFormat % (res.id.position, res.type, rotamer_Count)
    outputSurvivalFile.write(line_Out)
line_Out = "Total rotamer count: " + str(totalRotCount) + "   Total number of edges (not counting self-loops): " + str(edgeCount_G)
outputSurvivalFile.write(line_Out)

print "\nEliminating Rotamers:"
print "Res    Rotamer"
print "Pos:   Number:"

# Now do the DEE:
allRotamers_G.doGoldsteinLowEnergyRotamer()

print "End of first pass\n"
print "\nEliminating More Rotamers:"
print "Res    Rotamer"
print "Pos:   Number:"

allRotamers_G.doGoldsteinAllRotamersVsAllRotamers()

# Report results:
print "\nFINAL RESULTS after Dead End Elimination:"
print "Line    Res  Rot   Res  Rot    Pair"
print "Count:  Pos: Num:  Pos: Num:   Energy:\n"

printFormat = '%5d  %4d  %3d  %4d  %3d  %8.4f'
lineCount = 0
for res in allRotamers_G.siteRotamers_D.keys():
    for rotNum in allRotamers_G.siteRotamers_D[res].singleRotamers_D.keys():
        lineCount += 1
        self_energy = allRotamers_G.siteRotamers_D[res].singleRotamers_D[rotNum].selfEnergy
        lineOut = outFormat % (lineCount, res.id.position, rotNum, res.id.position, rotNum, self_energy)
        outputFile.write(lineOut)
        print printFormat % (lineCount, res.id.position, rotNum, res.id.position, rotNum, self_energy)
        for nearbyRes in nbrsMap_D_G[res]:
            if not nearbyRes in residuesWithRotamers_L_G: continue
            if nearbyRes == res: continue
            if nearbyRes.id.position < res.id.position: continue
            for nearbyRotNum in allRotamers_G.siteRotamers_D[nearbyRes].singleRotamers_D.keys():
                e = rotamerPairEnergy_D_G[res.id.position, rotNum, nearbyRes.id.position, nearbyRotNum]
                if abs(e) < 0.000001: continue
                lineCount += 1
                lineOut = outFormat % (lineCount, res.id.position, rotNum, nearbyRes.id.position, nearbyRotNum, e)
                outputFile.write(lineOut)
                print printFormat % (lineCount, res.id.position, rotNum, nearbyRes.id.position, nearbyRotNum, e)
outputFile.close()

# Record rotamer statistics after DEE:
outputSurvivalFile.write("\n\nAfter DEE:\n")
totalRotCount = 0
for res in allRotamers_G.siteRotamers_D.keys():
    rotamer_Count = len(allRotamers_G.siteRotamers_D[res].singleRotamers_D.keys())
    totalRotCount += rotamer_Count
    line_Out = outSurvivalFormat % (res.id.position, res.type, rotamer_Count)
    outputSurvivalFile.write(line_Out)
line_Out = "Total rotamer count: " + str(totalRotCount)+ "    Total number of edges (including self-loops): " + str(lineCount)
outputSurvivalFile.write(line_Out)
outputSurvivalFile.close()


# How many of the surviving rotamers have chi1 and chi2 within 40 degrees of the original (PDB) values?
numberOfChi1angles = 0
numberOfChi2angles = 0
numberChi1Within40 = 0
numberChi2Within40 = 0
print " Res  Rew   Rot    Within 40"
print " Pos: Type: Num:   Chi1: Chi2:"
survFormat = "%4d   %3s %3d     %6s    %6s"
for res in allRotamers_G.siteRotamers_D.keys():
    for rotix in allRotamers_G.siteRotamers_D[res].singleRotamers_D.keys():
        within40_Chi1 = allRotamers_G.siteRotamers_D[res].singleRotamers_D[rotix].within40Chi1
        if within40_Chi1 != None:
            numberOfChi1angles += 1
            if within40_Chi1: numberChi1Within40 += 1
        within40_Chi2 = allRotamers_G.siteRotamers_D[res].singleRotamers_D[rotix].within40Chi2
        if within40_Chi2 != None:
            numberOfChi2angles += 1
            if within40_Chi2: numberChi2Within40 += 1
        print survFormat % (res.id.position, res.type, rotix, within40_Chi1, within40_Chi2)

print "Percentage of chi1 angles within 40 degrees of original (PDB) chi1: ", 100.0* float(numberChi1Within40)/float(numberOfChi1angles)
print "Percentage of chi2 angles within 40 degrees of original (PDB) chi2: ", 100.0* float(numberChi2Within40)/float(numberOfChi2angles)
