import chimera, numpy, sets
from chimera import runCommand
from numpy import array, linalg
from sets import Set

from StructBio.StructureOverlap.overlapper import Overlapper
from StructBio.Utilities.shellNeighbors import Shell
from StructBio.Utilities.miscFunctions import userSpecificationOfChains, colorByName
from StructBio.Utilities.miscFunctions import residueKDhColor, residueFletterickColor
from StructBio.StructureOverlap.sequencesLocalAlignment import SequencesLocalAlignment
from StructBio.Graphs.graph import Graph, Node, Edge

from crn import RRcontacts, ContactRearrangementNetwork

#------------------------------------------------------------------------------
def getStandardResidues(prot):
    r_L = []
    for r in prot.residues:
        if r.type in ["ALA", "ARG", "ASN", "ASP", "CYS",
                      "GLU", "GLN", "GLY", "HIS", "ILE",
                      "LEU", "LYS", "MET", "PHE", "PRO",
                      "SER", "THR", "TRP", "TYR", "VAL"] and r.findAtom('CA'): r_L.append(r)
    return r_L

#------------------------------------------------------------------------------
def residueColor(res):
#    return residueKDhColor(res)
    return residueFletterickColor(res)    

#==============================================================================
# Mainline
#==============================================================================
runCommand("set bg_color white")

# Atom counts (side chain only) used to determine radii of nodes in graph:
atomCounts_D = {"ALA":1, "ARG":7, "ASN":4, "ASP":4, "CYS":2,
                "GLU":5, "GLN":5, "GLY":0, "HIS":6, "ILE":4,
                "LEU":4, "LYS":5, "MET":4, "PHE":7, "PRO":3,
                "SER":2, "THR":3, "TRP":10, "TYR":8, "VAL":3}

(protI, protA) = userSpecificationOfChains(2)

# Ideally both chains should have the same primary sequence but this might not guaranteed!
# It is possible that one chain has missing residues that do occur in the other chain.
# It is possible that one chain has one or more residue mutations! (Consider 1E0S and 1HFV).
# So, it is necessary to do a sequence alignment working with standard residues each
# containing an alpha carbon. This last requirement seems strange but is needed to get
# past a PDB file with an erroneous residue entry...

rI_L = getStandardResidues(protI)
rA_L = getStandardResidues(protA)
alignedPairs = SequencesLocalAlignment(rI_L, rA_L)
[rI_L, rA_L] = zip(*alignedPairs.alignment_L)

# Get rid of residues that do not match:  
residuesI_L = []
residuesA_L = []
for ix in range(len(rI_L)):
    if rI_L[ix].type == rA_L[ix].type:
        residuesI_L.append(rI_L[ix])
        residuesA_L.append(rA_L[ix])
    
crn = ContactRearrangementNetwork(residuesI_L, residuesA_L, 5.0, 0.3)

# Overlap the protein chains:
resInOverlap_L = []
for ix in range(len(crn.rrContacts_L)):
    resInOverlap_L.append((residuesI_L[crn.rrContacts_L[ix].resXix].findAtom("CA"),
                           residuesA_L[crn.rrContacts_L[ix].resYix].findAtom("CA")))

overlapper = Overlapper(protI)
overlapper.moveToOverlap(resInOverlap_L, massWeighted = True)
runCommand("focus")

#raw_input("Press Enter to continue.") # Use this to see the improvement in overlap.
# Do another overlap using 1/3 of the alpha carbon atoms (picking closest pairs).
closePairs_L = []
for ix in range(len(residuesI_L)):
    diff = linalg.norm(residuesI_L[ix].findAtom("CA").coord() - residuesA_L[ix].findAtom("CA").coord())
    closePairs_L.append((diff, (residuesI_L[ix].findAtom("CA"), residuesA_L[ix].findAtom("CA"))))
closePairs_L.sort()
overlapper = Overlapper(protI)
resInOverlap_L = []
for ix in range(len(residuesI_L)/3):
    resInOverlap_L.append(closePairs_L[ix][1])
overlapper.moveToOverlap(resInOverlap_L, massWeighted = True)


# Build graphs for both proteins:
crnGraphProtI = Graph(graphName = "Inactive State")
crnGraphProtA = Graph(graphName = "Active State")

for r in residuesA_L:
    nodeRadius = float(atomCounts_D[r.type] + 1)/10.
    crnGraphProtA.addNode(r.type + str(r.id.position),
                          Node(0.0, array(r.findAtom("CA").labelCoord()), nodeRadius, residueColor(r)))
for r in residuesI_L:
    nodeRadius = float(atomCounts_D[r.type] + 1)/10.
    crnGraphProtI.addNode(r.type + str(r.id.position),
                          Node(0.0, array(r.findAtom("CA").labelCoord()), nodeRadius, residueColor(r)))
    
for rrContact in crn.rrContacts_L:
    # Use the next statement for a graph that represents "sustained contact" during the state transition.
    #if rrContact.numAcontacts == 0 or rrContact.numIcontacts == 0: continue
    rA = residuesA_L[rrContact.resXix]
    nA = residuesA_L[rrContact.resYix] 
    rALabel = rA.type + str(rA.id.position)
    nALabel = nA.type + str(nA.id.position)
    rI = residuesI_L[rrContact.resXix]
    nI = residuesI_L[rrContact.resYix] 
    rILabel = rI.type + str(rI.id.position)
    nILabel = nI.type + str(nI.id.position)    
    spndRadius = abs(rrContact.numAcontacts - rrContact.numIcontacts)/25.0
    if rrContact.numAcontacts >= rrContact.numIcontacts:
        crnGraphProtI.addEdge(rILabel, nILabel, Edge(0.0, spndRadius, colorByName("light green")))
        crnGraphProtA.addEdge(rALabel, nALabel, Edge(0.0, spndRadius, colorByName("deep pink")))
    else:
        crnGraphProtA.addEdge(rALabel, nALabel, Edge(0.0, spndRadius, colorByName("light green")))
        crnGraphProtI.addEdge(rILabel, nILabel, Edge(0.0, spndRadius, colorByName("deep pink")))
        
crnGraphProtI.solids.display()
crnGraphProtA.solids.display()        

connectedComponentsStateI_L = crnGraphProtI.getConnectedComponents(3, "Connected component Inactive state")
for cc in connectedComponentsStateI_L:
    cc.solids.display()

connectedComponentsStateA_L = crnGraphProtA.getConnectedComponents(3, "Connected component Active state")
for cc in connectedComponentsStateA_L:
    cc.solids.display()

runCommand("focus")

















